@@ -18,7 +18,7 @@
 using namespace Okular;
 
 FormFieldPrivate::FormFieldPrivate( FormField::FieldType type )
-    : m_type( type ), m_activateAction( 0 )
+    : m_type( type ), m_activateAction( nullptr )
 {
 }
 
@@ -55,6 +55,10 @@ bool FormField::isReadOnly() const
     return false;
 }
 
+void FormField::setReadOnly( bool )
+{
+}
+
 bool FormField::isVisible() const
 {
     return true;
