@@ -517,6 +517,9 @@ PDFGenerator::PDFGenerator( QObject *parent, const QVariantList &args )
     setFeature( ReadRawData );
     setFeature( TiledRendering );
     setFeature( SwapBackingFile );
+#ifdef HAVE_POPPLER_0_63
+    setFeature( SupportsCancelling );
+#endif
 
     // You only need to do it once not for each of the documents but it is cheap enough
     // so doing it all the time won't hurt either
@@ -910,9 +913,9 @@ bool PDFGenerator::isAllowed( Okular::Permission permission ) const
 }
 
 #ifdef HAVE_POPPLER_0_62
-struct PartialUpdatePayload
+struct RenderImagePayload
 {
-    PartialUpdatePayload(PDFGenerator *g, Okular::PixmapRequest *r) :
+    RenderImagePayload(PDFGenerator *g, Okular::PixmapRequest *r) :
         generator(g), request(r)
     {
         // Don't report partial updates for the first 500 ms
@@ -925,11 +928,11 @@ struct PartialUpdatePayload
     Okular::PixmapRequest *request;
     QTimer timer;
 };
-Q_DECLARE_METATYPE(PartialUpdatePayload*)
+Q_DECLARE_METATYPE(RenderImagePayload*)
 
 static bool shouldDoPartialUpdateCallback(const QVariant &vPayload)
 {
-    auto payload = vPayload.value<PartialUpdatePayload *>();
+    auto payload = vPayload.value<RenderImagePayload *>();
 
     // Since the timer lives in a thread without an event loop we need to stop it ourselves
     // when the remaining time has reached 0
@@ -942,11 +945,19 @@ static bool shouldDoPartialUpdateCallback(const QVariant &vPayload)
 
 static void partialUpdateCallback(const QImage &image, const QVariant &vPayload)
 {
-    auto payload = vPayload.value<PartialUpdatePayload *>();
+    auto payload = vPayload.value<RenderImagePayload *>();
     QMetaObject::invokeMethod(payload->generator, "signalPartialPixmapRequest", Qt::QueuedConnection, Q_ARG(Okular::PixmapRequest*, payload->request), Q_ARG(QImage, image));
 }
 #endif
 
+#ifdef HAVE_POPPLER_0_63
+static bool shouldAbortRenderCallback(const QVariant &vPayload)
+{
+    auto payload = vPayload.value<RenderImagePayload *>();
+    return payload->request->shouldAbortRender();
+}
+#endif
+
 QImage PDFGenerator::image( Okular::PixmapRequest * request )
 {
     // debug requests to this (xpdf) generator
@@ -970,6 +981,12 @@ QImage PDFGenerator::image( Okular::PixmapRequest * request )
     // 0. LOCK [waits for the thread end]
     userMutex()->lock();
 
+    if ( request->shouldAbortRender() )
+    {
+        userMutex()->unlock();
+        return QImage();
+    }
+
     // 1. Set OutputDev parameters and Generate contents
     // note: thread safety is set on 'false' for the GUI (this) thread
     Poppler::Page *p = pdfdoc->page(page->number());
@@ -978,37 +995,79 @@ QImage PDFGenerator::image( Okular::PixmapRequest * request )
     QImage img;
     if (p)
     {
+#ifdef HAVE_POPPLER_0_63
         if ( request->isTile() )
         {
-            QRect rect = request->normalizedRect().geometry( request->width(), request->height() );
-#ifdef HAVE_POPPLER_0_62
+            const QRect rect = request->normalizedRect().geometry( request->width(), request->height() );
             if ( request->partialUpdatesWanted() )
             {
-                PartialUpdatePayload payload( this, request );
+                RenderImagePayload payload( this, request );
+                img = p->renderToImage( fakeDpiX, fakeDpiY, rect.x(), rect.y(), rect.width(), rect.height(), Poppler::Page::Rotate0,
+                                        partialUpdateCallback, shouldDoPartialUpdateCallback, shouldAbortRenderCallback, QVariant::fromValue( &payload ) );
+            }
+            else
+            {
+                RenderImagePayload payload( this, request );
+                img = p->renderToImage( fakeDpiX, fakeDpiY, rect.x(), rect.y(), rect.width(), rect.height(), Poppler::Page::Rotate0,
+                                        nullptr, nullptr, shouldAbortRenderCallback, QVariant::fromValue( &payload ) );
+            }
+        }
+        else
+        {
+            if ( request->partialUpdatesWanted() )
+            {
+                RenderImagePayload payload( this, request );
+                img = p->renderToImage( fakeDpiX, fakeDpiY, -1, -1, -1, -1, Poppler::Page::Rotate0,
+                                        partialUpdateCallback, shouldDoPartialUpdateCallback, shouldAbortRenderCallback, QVariant::fromValue( &payload ) );
+            }
+            else
+            {
+                RenderImagePayload payload( this, request );
+                img = p->renderToImage( fakeDpiX, fakeDpiY, -1, -1, -1, -1, Poppler::Page::Rotate0,
+                                        nullptr, nullptr, shouldAbortRenderCallback, QVariant::fromValue( &payload ) );
+            }
+
+        }
+#elif defined(HAVE_POPPLER_0_62)
+        if ( request->isTile() )
+        {
+            const QRect rect = request->normalizedRect().geometry( request->width(), request->height() );
+            if ( request->partialUpdatesWanted() )
+            {
+                RenderImagePayload payload( this, request );
                 img = p->renderToImage( fakeDpiX, fakeDpiY, rect.x(), rect.y(), rect.width(), rect.height(), Poppler::Page::Rotate0,
                                         partialUpdateCallback, shouldDoPartialUpdateCallback, QVariant::fromValue( &payload ) );
             }
             else
-#endif
             {
                 img = p->renderToImage( fakeDpiX, fakeDpiY, rect.x(), rect.y(), rect.width(), rect.height(), Poppler::Page::Rotate0 );
             }
         }
         else
         {
-#ifdef HAVE_POPPLER_0_62
             if ( request->partialUpdatesWanted() )
             {
-                PartialUpdatePayload payload(this, request);
+                RenderImagePayload payload( this, request );
                 img = p->renderToImage( fakeDpiX, fakeDpiY, -1, -1, -1, -1, Poppler::Page::Rotate0,
                                         partialUpdateCallback, shouldDoPartialUpdateCallback, QVariant::fromValue( &payload ) );
             }
             else
-#endif
             {
-                img = p->renderToImage(fakeDpiX, fakeDpiY, -1, -1, -1, -1, Poppler::Page::Rotate0 );
+                img = p->renderToImage( fakeDpiX, fakeDpiY, -1, -1, -1, -1, Poppler::Page::Rotate0 );
             }
+
         }
+#else
+        if ( request->isTile() )
+        {
+            const QRect rect = request->normalizedRect().geometry( request->width(), request->height() );
+            img = p->renderToImage( fakeDpiX, fakeDpiY, rect.x(), rect.y(), rect.width(), rect.height(), Poppler::Page::Rotate0 );
+        }
+        else
+        {
+            img = p->renderToImage( fakeDpiX, fakeDpiY, -1, -1, -1, -1, Poppler::Page::Rotate0 );
+        }
+#endif
     }
     else
     {
@@ -1100,8 +1159,28 @@ void PDFGenerator::resolveMediaLinkReferences( Okular::Page *page )
         resolveMediaLinkReference( field->activationAction() );
 }
 
-Okular::TextPage* PDFGenerator::textPage( Okular::Page *page )
+#ifdef HAVE_POPPLER_0_63
+struct TextExtractionPayload
 {
+    TextExtractionPayload(Okular::TextRequest *r) :
+        request(r)
+    {
+    }
+
+    Okular::TextRequest *request;
+};
+Q_DECLARE_METATYPE(TextExtractionPayload*)
+
+static bool shouldAbortTextExtractionCallback(const QVariant &vPayload)
+{
+    auto payload = vPayload.value<TextExtractionPayload *>();
+    return payload->request->shouldAbortExtraction();
+}
+#endif
+
+Okular::TextPage* PDFGenerator::textPage( Okular::TextRequest *request )
+{
+    const Okular::Page *page = request->page();
 #ifdef PDFGENERATOR_DEBUG
     qCDebug(OkularPdfDebug) << "page" << page->number();
 #endif
@@ -1112,7 +1191,12 @@ Okular::TextPage* PDFGenerator::textPage( Okular::Page *page )
     Poppler::Page *pp = pdfdoc->page( page->number() );
     if (pp)
     {
+#ifdef HAVE_POPPLER_0_63
+        TextExtractionPayload payload(request);
+        textList = pp->textList( Poppler::Page::Rotate0, shouldAbortTextExtractionCallback, QVariant::fromValue( &payload ) );
+#else
         textList = pp->textList();
+#endif
         const QSizeF s = pp->pageSizeF();
         pageWidth = s.width();
         pageHeight = s.height();
@@ -1125,6 +1209,9 @@ Okular::TextPage* PDFGenerator::textPage( Okular::Page *page )
     delete pp;
     userMutex()->unlock();
 
+    if ( textList.isEmpty() && request->shouldAbortExtraction() )
+        return nullptr;
+
     Okular::TextPage *tp = abstractTextPage(textList, pageHeight, pageWidth, (Poppler::Page::Rotation)page->orientation());
     qDeleteAll(textList);
     return tp;
